# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2015, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import apsw
import contextlib
import numpy.random
import random
import struct

import bayeslite.bql as bql
import bayeslite.bqlfn as bqlfn
import bayeslite.metamodel as metamodel
import bayeslite.parse as parse
import bayeslite.schema as schema
import bayeslite.txn as txn
import bayeslite.weakprng as weakprng

from bayeslite.util import cursor_value

bayesdb_open_cookie = 0xed63e2c26d621a5b5146a334849d43f0

def bayesdb_open(pathname=None, builtin_metamodels=None, seed=None,
        version=None, compatible=None):
    """Open the BayesDB in the file at `pathname`.

    If there is no file at `pathname`, it is automatically created.
    If `pathname` is unspecified or ``None``, a temporary in-memory
    BayesDB instance is created.

    `seed` is a 32-byte string specifying a pseudorandom number
    generation seed.  If not specified, it defaults to all zeros.

    If `compatible` is `None` or `False` and the database already
    exists, `bayesdb_open` may have the effect of incompatibly
    changing the format of the database so that older versions of
    bayeslite cannot read it.  If `compatible` is `True`,
    `bayesdb_open` will not incompatibly change the format of the
    database.
    """
    if builtin_metamodels is None:
        builtin_metamodels = True
    bdb = BayesDB(bayesdb_open_cookie, pathname=pathname, seed=seed,
        version=version, compatible=compatible)
    if builtin_metamodels:
        metamodel.bayesdb_register_builtin_metamodels(bdb)
    return bdb

class BayesDB(object):
    """A handle for a Bayesian database in memory or on disk.

    Do not create BayesDB instances directly; use :func:`bayesdb_open` instead.

    An instance of `BayesDB` is a context manager that returns itself
    on entry and closes itself on exit, so you can write::

        with bayesdb_open(pathname='foo.bdb') as bdb:
            ...
    """

    def __init__(self, cookie, pathname=None, seed=None, version=None,
            compatible=None):
        if cookie != bayesdb_open_cookie:
            raise ValueError('Do not construct BayesDB objects directly!')
        if pathname is None:
            pathname = ":memory:"
        self._sqlite3 = apsw.Connection(pathname)
        self.txn_depth = 0
        self.metamodels = {}
        self.tracer = None
        self.sql_tracer = None
        self.cache = None
        self.temptable = 0
        self.qid = 0
        if seed is None:
            seed = struct.pack('<QQQQ', 0, 0, 0, 0)
        self._prng = weakprng.weakprng(seed)
        pyrseed = self._prng.weakrandom32()
        self._py_prng = random.Random(pyrseed)
        nprseed = [self._prng.weakrandom32() for _ in range(4)]
        self._np_prng = numpy.random.RandomState(nprseed)
        schema.bayesdb_install_schema(self._sqlite3, version=version,
            compatible=compatible)
        bqlfn.bayesdb_install_bql(self._sqlite3, self)

        # Cache an empty cursor for convenience.
        empty_cursor = self._sqlite3.cursor()
        empty_cursor.execute('')
        self._empty_cursor = bql.BayesDBCursor(self, empty_cursor)

    def __enter__(self):
        return self
    def __exit__(self, *_exc_info):
        self.close()

    def close(self):
        """Close the database.  Further use is not allowed."""
        assert self.txn_depth == 0, "pending BayesDB transactions"
        self._sqlite3.close()
        self._sqlite3 = None

    @property
    def py_prng(self):
        """A :class:`random.Random` object local to this BayesDB instance.

        This pseudorandom number generator is deterministically
        initialized from the seed supplied to :func:`bayesdb_open`.
        Use it to conserve reproducibility of results.
        """
        return self._py_prng

    @property
    def np_prng(self):
        """A Numpy RandomState object local to this BayesDB instance.

        This pseudorandom number generator is deterministically
        initialized from the seed supplied to :func:`bayesdb_open`.
        Use it to conserve reproducibility of results.
        """
        return self._np_prng

    def trace(self, tracer):
        """Trace execution of BQL queries.

        For simple tracing, pass a function or arbitrary Python
        callable as the `tracer`.  It will be called at the start of
        execution of each BQL query, with two arguments: the query to
        be executed, as a string; and the sequence or dictionary of
        bindings.

        For articulated tracing, pass an instance of
        :class:`~IBayesDBTracer`, whose methods will be called in the
        pattern described in its documentation.

        Only one tracer can be established at a time.  To remove it,
        use :meth:`~BayesDB.untrace`.

        """
        assert self.tracer is None
        self.tracer = tracer

    def untrace(self, tracer):
        """Stop tracing execution of BQL queries.

        `tracer` must have been previously established with
        :meth:`~BayesDB.trace`.

        Any queries currently in progress will continue to be traced
        until completion.

        """
        assert self.tracer == tracer
        self.tracer = None

    def sql_trace(self, tracer):
        """Trace execution of SQL queries.

        For simple tracing, pass a function or arbitrary Python
        callable as the `tracer`.  It will be called at the start of
        execution of each SQL query, with two arguments: the query to
        be executed, as a string; and the sequence or dictionary of
        bindings.

        For articulated tracing, pass an instance of
        :class:`~IBayesDBTracer`, whose methods will be called in the
        pattern described in its documentation.

        Only one tracer can be established at a time.  To remove it,
        use :meth:`~BayesDB.sql_untrace`.

        """
        assert self.sql_tracer is None
        self.sql_tracer = tracer

    def sql_untrace(self, tracer):
        """Stop tracing execution of SQL queries.

        `tracer` must have been previously established with
        :meth:`~BayesDB.sql_trace`.

        Any queries currently in progress will continue to be traced
        until completion.

        """
        assert self.sql_tracer == tracer
        self.sql_tracer = None

    def execute(self, string, bindings=None):
        """Execute a BQL query and return a cursor for its results.

        The argument `string` is a string parsed into a single BQL
        query.  It must contain exactly one BQL phrase, optionally
        terminated by a semicolon.

        The argument `bindings` is a sequence or dictionary of
        bindings for parameters in the query, or ``None`` to supply no
        bindings.
        """
        if bindings is None:
            bindings = ()
        return self._maybe_trace(
            self.tracer, self._do_execute, string, bindings)

    def _maybe_trace(self, tracer, meth, string, bindings):
        if tracer and isinstance(tracer, IBayesDBTracer):
            return self._trace_articulately(
                tracer, meth, string, bindings)
        if tracer:
            tracer(string, bindings)
        return meth(string, bindings)

    def _qid(self):
        self.qid += 1
        return self.qid

    def _trace_articulately(self, tracer, meth, string, bindings):
        qid = self._qid()
        tracer.start(qid, string, bindings)
        try:
            cursor = meth(string, bindings)
            tracer.ready(qid, cursor)
            if cursor == self._empty_cursor:
                tracer.finished(qid)
                # Calling abandoned here is a choice.  On the one
                # hand, I know the client can't get anything out of
                # the cursor; on the other hand, returning a
                # TracingCursor would let the tracer detect when the
                # client had released their null result cursor.  Why
                # would the tracer care about that?
                tracer.abandoned(qid)
                return cursor
            else:
                return TracingCursor(tracer, qid, cursor)
        except Exception as e:
            tracer.error(qid, e)
            raise

    def _do_execute(self, string, bindings):
        phrases = parse.parse_bql_string(string)
        phrase = None
        try:
            phrase = phrases.next()
        except StopIteration:
            raise ValueError('no BQL phrase in string')
        try:
            phrases.next()
        except StopIteration:
            pass
        else:
            raise ValueError('>1 phrase in string')
        cursor = bql.execute_phrase(self, phrase, bindings)
        return self._empty_cursor if cursor is None else cursor

    def sql_execute(self, string, bindings=None):
        """Execute a SQL query on the underlying SQLite database.

        The argument `string` is a string parsed into a single SQL
        query.  It must contain exactly one SQL phrase, optionally
        terminated by a semicolon.

        The argument `bindings` is a sequence or dictionary of
        bindings for parameters in the query, or ``None`` to supply no
        bindings.
        """
        if bindings is None:
            bindings = ()
        return self._maybe_trace(
            self.sql_tracer, self._do_sql_execute, string, bindings)

    def _do_sql_execute(self, string, bindings):
        cursor = self._sqlite3.cursor()
        cursor.execute(string, bindings)
        return bql.BayesDBCursor(self, cursor)

    @contextlib.contextmanager
    def savepoint(self):
        """Savepoint context.  On return, commit; on exception, roll back.

        Savepoints may be nested.  Parsed metadata and models are
        cached in Python during a savepoint.
        """
        with txn.bayesdb_savepoint(self):
            yield

    @contextlib.contextmanager
    def savepoint_rollback(self):
        """Auto-rollback savepoint context.  Roll back on return or exception.
        """
        with txn.bayesdb_savepoint_rollback(self):
            yield

    @contextlib.contextmanager
    def transaction(self):
        """Transaction context.  On return, commit; on exception, roll back.

        Transactions may not be nested: use a savepoint if you need
        nesting.  Parsed metadata and models are cached in Python
        during a savepoint.
        """
        with txn.bayesdb_transaction(self):
            yield

    def temp_table_name(self):
        n = self.temptable
        self.temptable += 1
        return 'bayesdb_temp_%u' % (n,)

    def last_insert_rowid(self):
        """Return the rowid of the row most recently inserted."""
        return self._sqlite3.last_insert_rowid()

class IBayesDBTracer(object):
    """BayesDB articulated tracing interface.

    If you just want to trace start of queries, pass a function to
    :meth:`~BayesDB.trace` or :meth:`~BayesDB.sql_trace`.  If you want
    finer-grained event tracing, pass an instance of this interface.

    A successful execution of a BayesDB query goes through the
    following stages:

    0. Not started
    1. Preparing cursor
    2. Result available
    3. All results consumed

    Preparing the cursor and consuming results may both be fast or
    slow, and may succeed or fail, depending on the query.  Also, the
    client may abandon some queries without consuming all the results.

    Thus, a query may experience the following transitions:

    - start: 0 --> 1
    - ready: 1 --> 2
    - error: 1 --> 0 or 2 --> 0
    - finished: 2 --> 3
    - abandoned: 2 --> 0 or 3 --> 0

    To receive notifications of any of those events for BQL or SQL
    queries, override the corresponding method(s) of this interface,
    and install the tracer object using :meth:`~BayesDB.trace` or
    :meth:`~BayesDB.sql_trace` respectively.

    Note 1: The client may run multiple cursors at the same time, so
    queries in the "Result available" state may overlap.

    Note 2: Abandonment of a query is detected when the cursor object
    is garbage collected, so the timing cannot be relied upon.

    """

    def start(self, qid, query, bindings):
        """Called when a query is started.

        The arguments are a unique query id, the query string, and the
        tuple or dictionary of bindings.

        """
        pass

    def ready(self, qid, cursor):
        """Called when a query is ready for consumption of results.

        The arguments are the query id and the BayesDB cursor.

        Note for garbage collector wonks: the passed `cursor` is the
        one wrapped in the :class:`~TracingCursor`, not the
        TracingCursor instance itself, so a tracer retaining a
        reference to `cursor` will not create a reference cycle or
        prevent the :meth:`abandoned` method from being called.

        """
        pass

    def error(self, qid, e):
        """Called when query preparation or result consumption fails.

        The arguments are the query id and the exception object.

        """
        pass

    def finished(self, qid):
        """Called when all query results are consumed."""
        pass

    def abandoned(self, qid):
        """Called when a query is abandoned.

        This is detected when the cursor object is garbage collected,
        so its timing cannot be relied upon.

        """
        pass

class TracingCursor(object):
    """Cursor wrapper for tracing interaction with an underlying cursor."""
    def __init__(self, tracer, qid, cursor):
        self._tracer = tracer
        self._qid = qid
        self._cursor = cursor

    def __iter__(self):
        return self

    def next(self):
        try:
            return self._cursor.next()
        except StopIteration:
            self._tracer.finished(self._qid)
            raise
        except Exception as e:
            self._tracer.error(self._qid, e)
            raise

    def fetchvalue(self):
        return cursor_value(self)

    def fetchone(self):
        try:
            ans = self._cursor.fetchone()
            if ans is None:
                self._tracer.finished(self._qid)
            return ans
        except Exception as e:
            self._tracer.error(self._qid, e)
            raise

    def fetchmany(self, size=1):
        try:
            ans = self._cursor.fetchmany(size=size)
            if len(ans) < size:
                self._tracer.finished(self._qid)
            return ans
        except Exception as e:
            self._tracer.error(self._qid, e)
            raise

    def fetchall(self):
        try:
            ans = self._cursor.fetchall()
            self._tracer.finished(self._qid)
            return ans
        except Exception as e:
            self._tracer.error(self._qid, e)
            raise

    @property
    def connection(self):
        return self._cursor.connection
    @property
    def lastrowid(self):
        return self._cursor.lastrowid
    @property
    def description(self):
        desc = self._cursor.description
        return [] if desc is None else desc

    def __del__(self):
        self._tracer.abandoned(self._qid)
        del self._tracer
        del self._qid
        del self._cursor



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2015, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Logging sessions.

PLEASE DO NOT SEND SESSIONS WHEN THEY MAY CONTAIN SECRET OR SENSITIVE INFO!

If you want to log your interactions with BayesDB in order to send a
record of all your queries to the MIT Probabilistic Computing Project
for diagnostic and research purposes, create a `SessionOrchestrator`
object:

    import bayeslite.sessions as sessions

    so = sessions.SessionOrchestrator(bdb)

This will begin recording in the database all queries submitted to
`bdb` with `bdb.execute` and `bdb.sql_execute`, including those done
internally by BayesDB -- specifically, it will record:

- the text of each query
- the parameters passed to each query
- the starting time of each query
- the ending time of each query, if it completed
- the stack trace of any error while executing a query
- the version of bayeslite used to create each session

To upload the session to the MIT Probcomp group over the internet, run:

    so.send_session_data()

Please contact probcomp-community@csail.mit.edu with questions and
concerns.

PLEASE DO NOT SEND SESSIONS WHEN THEY MAY CONTAIN SECRET OR SENSITIVE INFO!
"""

# PLEASE DO NOT SEND SESSIONS WHEN THEY MAY CONTAIN SECRET OR SENSITIVE INFO!
#
# For details on what is recorded in sessions and where, see:
#
# SessionOrchestrator._add_entry
#
#       Stores query text and parameters in the database.
#
# SessionOrchestrator.dump_session_as_json
#
#       Serializes all data about a session into a JSON blob.
#
# SessionOrchestrator.send_session_data
#
#       Sends session data over the internet in an HTTP POST request.
#
# See also the bayesdb_session and bayesdb_session_entries tables in
# the database schema, in schema.py.
#
# PLEASE DO NOT SEND SESSIONS WHEN THEY MAY CONTAIN SECRET OR SENSITIVE INFO!

import apsw
import json
import requests
import sys
import time
import traceback

import bayeslite

from bayeslite import IBayesDBTracer
from bayeslite.schema import bayesdb_schema_required
from bayeslite.util import cursor_value
from bayeslite import __version__

_error_previous_session_msg = 'WARNING: Current or previous session contains queries that resulted in errors or exceptions. Consider uploading the session with send_session_data() if your data are freely shareable (free of secret or personally identifiable info). Contact probcomp-community@csail.mit.edu for help or with questions.'

class SessionOrchestrator(object):

    def __init__(self, bdb, logger=None, post=None):
        bayesdb_schema_required(bdb, 7, 'sessions')
        if post is None:
            post = requests.post
        self.bdb = bdb
        self._qid_to_entry_id = {}
        self._sql_tracer = _SessionTracer("sql", self)
        self._bql_tracer = _SessionTracer("bql", self)
        self._post = post
        self.start_saving_sessions()
        self._suggested_send = False
        self._logger = logger
        self._start_new_session()

    def _info(self, msg):
        if self._logger:
            self._logger.info(msg)
        else:
            print msg

    def _warn(self, msg):
        if self._logger:
            self._logger.warn(msg)
        else:
            print msg

    def _error(self, msg):
        if self._logger:
            self._logger.error(msg)
        else:
            print msg

    def _sql(self, query, bindings=None):
        # Go through bdb._sqlite3.cursor().execute instead of
        # bdb.sql_execute to avoid hitting the tracer.
        if bindings == None:
            bindings = ()
        return self.bdb._sqlite3.cursor().execute(query, bindings)

    def _add_entry(self, qid, type, query, bindings):
        '''Save a session entry into the database. The entry is initially in
        the not-completed state. Return the new entry's id so that it can be
        set to completed when appropriate.'''
        # check for errors on this session and suggest if we haven't already
        if not self._suggested_send:
            self._check_error_entries(self.session_id)
        data = query + json.dumps(bindings)
        self._sql('''
            INSERT INTO bayesdb_session_entries
                (session_id, type, data, start_time)
                VALUES (?,?,?,?)
        ''', (self.session_id, type, data, time.time()))
        entry_id = cursor_value(self._sql('SELECT last_insert_rowid()'))
        self._qid_to_entry_id[qid] = entry_id

    def _mark_entry_completed(self, qid):
        entry_id = self._qid_to_entry_id[qid]
        self._sql('''
            UPDATE bayesdb_session_entries SET end_time = ? WHERE id = ?
        ''', (time.time(), entry_id))

    def _mark_entry_error(self, qid):
        entry_id = self._qid_to_entry_id[qid]
        self._sql('''
            UPDATE bayesdb_session_entries
                SET error = ?, end_time = ?
                WHERE id = ?
        ''', (traceback.format_exc(), time.time(), entry_id))

    def _start_new_session(self):
        self._sql('INSERT INTO bayesdb_session (version) VALUES (?)',
                  (__version__,))
        self.session_id = cursor_value(self._sql('SELECT last_insert_rowid()'))
        # check for errors on the previous session
        self._check_error_entries(self.session_id - 1)

    def _check_error_entries(self, session_id):
        '''Check if the previous session contains queries that resulted in
        errors and suggest sending the session'''
        error_entries = cursor_value(self._sql('''
            SELECT COUNT(*) FROM bayesdb_session_entries
                WHERE error IS NOT NULL AND session_id = ?
        ''', (session_id,)))
        # suggest sending sessions but don't suggest more than once
        if (error_entries > 0 and not self._suggested_send):
            self._warn(_error_previous_session_msg)
            self._suggested_send = True
        return error_entries

    def clear_all_sessions(self):
        self._sql('DELETE FROM bayesdb_session_entries')
        self._sql('DELETE FROM bayesdb_session')
        self._sql('''
            DELETE FROM sqlite_sequence
                WHERE name = 'bayesdb_session'
                OR name = 'bayesdb_session_entries'
        ''')
        self._start_new_session()

    def list_sessions(self):
        """Lists all saved sessions with the number of entries in each, and
        whether they were sent or not."""
        return self._sql('SELECT * FROM bayesdb_session')

    def current_session_id(self):
        """Returns the current integer session id."""
        return self.session_id

    def dump_session_as_json(self, session_id):
        """Returns a JSON string representing the list of SQL or BQL entries
        (e.g.  queries) executed within session `session_id`."""
        if session_id > self.session_id or session_id < 1:
            raise ValueError('No such session (%d)' % session_id)
        version = cursor_value(self._sql('''
            SELECT version FROM bayesdb_session
                WHERE id = ?
        ''', (session_id,)))
        cursor = self._sql('''
            SELECT * FROM bayesdb_session_entries
                WHERE session_id = ?
                ORDER BY start_time DESC
        ''', (session_id,))
        # XXX Get the description first because apsw cursors, for
        # whatever reason, don't let you get the description after
        # you've gotten all the results.
        # (see also bql.py BayesDBCursor.__init__)
        fields = []
        try:
            fields = [d[0] for d in cursor.description]
        except apsw.ExecutionCompleteError:
            pass # Probably no rows.
        entries = cursor.fetchall()
        session = {
            'entries': entries,
            'fields': fields,
            'version': version,
        }
        return json.dumps(session, sort_keys=True)

    def dump_current_session_as_json(self):
        """Returns a JSON string representing the current sesion (see
        `dump_session_as_json`)"""
        return self.dump_session_as_json(self.session_id)

    def send_session_data(self):
        """Send all saved session history. The session history will be used for
        research purposes. DO NOT SEND IF YOU ARE WORKING WITH CONFIDENTIAL,
        PROPRIETARY, IDENTIFYING, OR OTHERWISE SECRET INFO."""
        probcomp_url = 'https://projects.csail.mit.edu/probcomp/bayesdb/save_sessions.cgi'
        for id in range(1, self.session_id+1):
            self._info('Sending session %d to %s ...' % (id, probcomp_url))
            json_string = self.dump_session_as_json(id)
            self._info(json_string)
            data = {'session_json': json_string,
                    'User-Agent': 'bayeslite %s' % (__version__,)}
            r = self._post(probcomp_url, data=data)
            self._info('Response: %s' % (r.text,))

    def start_saving_sessions(self):
        self.bdb.trace(self._bql_tracer)
        self.bdb.sql_trace(self._sql_tracer)

    def stop_saving_sessions(self):
        self.bdb.untrace(self._bql_tracer)
        self.bdb.sql_untrace(self._sql_tracer)

class _SessionTracer(IBayesDBTracer):

    def __init__(self, type, orchestrator):
        self._type = type
        self._orchestrator = orchestrator

    def start(self, qid, query, bindings):
        self._orchestrator._add_entry(qid, self._type, query, bindings)

    def finished(self, qid):
        # TODO: currently appears unreliable, error is being used instead
        self._orchestrator._mark_entry_completed(qid)

    def error(self, qid, e):
        self._orchestrator._mark_entry_error(qid)



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2015, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from bayeslite.exception import BayesDBException
from bayeslite.sqlite3_util import sqlite3_transaction
from bayeslite.util import cursor_value

bayesdb_schema_5 = '''
PRAGMA user_version = 5;

CREATE TABLE bayesdb_metamodel (
	name		TEXT COLLATE NOCASE NOT NULL PRIMARY KEY,
	version		INTEGER NOT NULL
);

-- Statistics type.
CREATE TABLE bayesdb_stattype (
	name		TEXT COLLATE NOCASE NOT NULL PRIMARY KEY
);

INSERT INTO bayesdb_stattype VALUES ('categorical');
INSERT INTO bayesdb_stattype VALUES ('cyclic');
INSERT INTO bayesdb_stattype VALUES ('numerical');

CREATE TABLE bayesdb_column (
	tabname		TEXT COLLATE NOCASE NOT NULL,
	colno		INTEGER NOT NULL CHECK (0 <= colno),
	name		TEXT COLLATE NOCASE NOT NULL,
	shortname	TEXT,
	description	TEXT,
	PRIMARY KEY(tabname, colno),
	UNIQUE(tabname, name)
);

CREATE TABLE bayesdb_column_map (
	tabname		TEXT COLLATE NOCASE NOT NULL,
	colno		INTEGER NOT NULL,
	key		TEXT NOT NULL,
	value		TEXT NOT NULL,
	PRIMARY KEY(tabname, colno, key),
	FOREIGN KEY(tabname, colno) REFERENCES bayesdb_column(tabname, colno)
);

CREATE TABLE bayesdb_generator (
	-- We use AUTOINCREMENT so that generator id numbers don't get
	-- reused and are safe to hang onto outside a transaction.
	id		INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT
				CHECK (0 < id),
	name		TEXT COLLATE NOCASE NOT NULL UNIQUE,
	tabname		TEXT COLLATE NOCASE NOT NULL,
				-- REFERENCES sqlite_master(name)
	metamodel	INTEGER NOT NULL REFERENCES bayesdb_metamodel(name)
);

CREATE TABLE bayesdb_generator_column (
	generator_id	INTEGER NOT NULL REFERENCES bayesdb_generator(id),
	colno		INTEGER NOT NULL,
	stattype	TEXT NOT NULL REFERENCES bayesdb_stattype(name),
	PRIMARY KEY(generator_id, colno)
);

CREATE TABLE bayesdb_generator_model (
	generator_id	INTEGER NOT NULL REFERENCES bayesdb_generator(id),
	modelno		INTEGER NOT NULL,
	iterations	INTEGER NOT NULL CHECK (0 <= iterations),
	PRIMARY KEY(generator_id, modelno)
);
'''

bayesdb_schema_5to6 = '''
PRAGMA user_version = 6;

ALTER TABLE bayesdb_generator
    ADD COLUMN defaultp BOOLEAN DEFAULT 0;

CREATE UNIQUE INDEX bayesdb_generator_i_default ON bayesdb_generator (tabname)
    WHERE defaultp;
'''

bayesdb_schema_6to7 = '''
PRAGMA user_version = 7;

CREATE TABLE bayesdb_session (
	id		INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT
				CHECK (0 < id),
	sent	BOOLEAN DEFAULT 0,
	version	TEXT NOT NULL
);

CREATE TABLE bayesdb_session_entries (
	id		INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT
				CHECK (0 < id),
	session_id	INTEGER NOT NULL REFERENCES bayesdb_session(id),
	type		TEXT CHECK (type IN ('bql','sql')) NOT NULL,
	data		TEXT NOT NULL,
	-- Timing is by the local POSIX clock.
	start_time	INTEGER NOT NULL,
	end_time	INTEGER,
	error		TEXT
);
'''

### BayesDB SQLite setup

def bayesdb_install_schema(db, version=None, compatible=None):
    # Get the application id.
    cursor = db.cursor().execute('PRAGMA application_id')
    application_id = 0
    try:
        row = cursor.next()
    except StopIteration:
        raise EnvironmentError('Missing application_id in sqlite3.')
    else:
        application_id = row[0]
        assert isinstance(application_id, int)

    # Get the user version.
    cursor = db.cursor().execute('PRAGMA user_version')
    user_version = 0
    try:
        row = cursor.next()
    except StopIteration:
        raise EnvironmentError('Missing user_version in sqlite3.')
    else:
        user_version = row[0]
        assert isinstance(user_version, int)

    # Check them.  If zero, install schema.  If mismatch, fail.
    install = False
    if application_id == 0 and user_version == 0:
        # Assume we just created the database.
        #
        # XXX What if we opened some random other sqlite file which
        # did not have an application_id or user_version set?  Hope
        # everyone else sets application_id and user_version too...
        with sqlite3_transaction(db):
            db.cursor().execute('PRAGMA application_id = %d' % (0x42594442,))
            db.cursor().execute(bayesdb_schema_5)
        user_version = 5
        install = True
    elif application_id != 0x42594442:
        raise IOError('Wrong application id: 0x%08x' % (application_id,))
    # 5 was the last prerelease version.
    if user_version == 5:
        with sqlite3_transaction(db):
            db.cursor().execute(bayesdb_schema_5to6)
        user_version = 6
    if user_version != 6 and user_version != 7:
        raise IOError('Unknown bayeslite format version: %d' % (user_version,))
    if install or not compatible:
        _upgrade_schema(db, version=version)
    db.cursor().execute('PRAGMA foreign_keys = ON')
    db.cursor().execute('PRAGMA integrity_check')
    db.cursor().execute('PRAGMA foreign_key_check')

def _upgrade_schema(db, version=None):
    if version is None:
        version = 7
    assert 6 <= version
    current_version = _schema_version(db)
    assert 6 <= current_version
    if current_version == 6 < version:
        with sqlite3_transaction(db):
            db.cursor().execute(bayesdb_schema_6to7)
        current_version = 7
    assert current_version == version
    db.cursor().execute('PRAGMA integrity_check')
    db.cursor().execute('PRAGMA foreign_key_check')

def _schema_version(db):
    return cursor_value(db.cursor().execute('PRAGMA user_version'))

def bayesdb_upgrade_schema(bdb, version=None):
    """Upgrade the BayesDB internal database schema.

    If `version` is `None`, upgrade to the latest database format
    version supported by bayeslite.  Otherwise, it may be a schema
    version number.
    """
    _upgrade_schema(bdb._sqlite3, version=version)

def bayesdb_schema_version(bdb):
    """Return the version number for the BayesDB internal database schema."""
    return _schema_version(bdb._sqlite3)

def bayesdb_schema_required(bdb, version, why):
    """Fail if `bdb`'s internal database schema version is not new enough.

    The string `why` is included in the text of the error message,
    along with advice to use `bayesdb_upgrade_schema` to upgrade it.
    """
    current_version = bayesdb_schema_version(bdb)
    if current_version < version:
        raise BayesDBException(bdb, 'BayesDB schema version too old'
            ' to support %s: current version %r, at least %r required;'
            ' use bayesdb_upgrade_schema to upgrade.'
            % (why, current_version, version))



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2015, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import apsw
import json
import pytest

from collections import defaultdict, namedtuple

import bayeslite
import bayeslite.metamodels.troll_rng as troll
import bayeslite.sessions as sescap
from bayeslite.version import __version__

from bayeslite.util import cursor_value

import test_core

def make_bdb():
    crosscat = test_core.local_crosscat()
    metamodel = test_core.CrosscatMetamodel(crosscat)
    bdb = bayeslite.bayesdb_open(builtin_metamodels=False)
    bayeslite.bayesdb_register_metamodel(bdb, metamodel)
    return bdb

def make_bdb_with_sessions(*args, **kwargs):
    bdb = make_bdb()
    tr = sescap.SessionOrchestrator(bdb, *args, **kwargs)
    return (bdb, tr)

def get_num_sessions(executor):
    return cursor_value(executor('''SELECT COUNT(*) FROM bayesdb_session'''))

def get_session_versions(executor):
    return list(executor('''SELECT version from bayesdb_session'''))

def get_num_entries(executor):
    return cursor_value(executor('''
        SELECT COUNT(*) FROM bayesdb_session_entries
    '''))

def get_entries(executor):
    entries = executor('''
        SELECT * FROM bayesdb_session_entries ORDER BY id
    ''')
    fields = [
        'id',
        'session_id',
        'type',
        'data',
        'start_time',
        'end_time',
        'error',
    ]
    assert fields == [d[0] for d in entries.description]
    SessionEntry = namedtuple('SessionEntry', fields)
    return [SessionEntry(*e) for e in entries]

def _basic_test_trace(executor):

    # a new session is automatically initialized with id 1
    assert get_num_sessions(executor) == 1

    # the above select query counts should become one or more entries
    num_entries = cursor_value(executor('''
        SELECT COUNT(*) FROM bayesdb_session_entries
    '''))
    assert num_entries > 0

    # entries are ordered starting from 1
    for id, entry in enumerate(get_entries(executor)):
        assert entry.session_id == 1
        assert entry.id == id + 1

def test_sessions_basic_bql():
    (bdb, tr) = make_bdb_with_sessions()
    _basic_test_trace(bdb.execute)

def test_sessions_basic_sql():
    (bdb, tr) = make_bdb_with_sessions()
    _basic_test_trace(bdb.sql_execute)

def _simple_bql_query(bdb):
    bdb.execute('''SELECT COUNT(*) FROM bayesdb_session''').fetchall()

def test_sessions_session_id_and_clear_sessions():
    (bdb, tr) = make_bdb_with_sessions()
    _simple_bql_query(bdb)

    # create two more sessions
    tr._start_new_session()
    tr._start_new_session()
    assert tr.current_session_id() == 3
    assert get_num_sessions(bdb.execute) == 3
    versions = get_session_versions(bdb.execute)
    assert 3 == len(versions)
    for version_row in versions:
        assert 1 == len(version_row)
        assert version_row[0] == __version__

    # there should now be one session (the current session)
    tr.clear_all_sessions()
    assert tr.current_session_id() == 1
    assert get_num_sessions(bdb.execute) == 1

    # the entry ids in the current session should start from 1
    assert min(entry.id for entry in get_entries(bdb.execute)) == 1

def test_sessions_start_stop():
    bdb = make_bdb()

    # the session table exists but there should be no entries before we
    # register the session tracer
    assert get_num_sessions(bdb.execute) == 0
    _simple_bql_query(bdb)
    assert get_num_entries(bdb.execute) == 0

    # registering the tracer starts recording of sessions
    tr = sescap.SessionOrchestrator(bdb)
    _simple_bql_query(bdb)
    num = get_num_entries(bdb.execute)
    assert num > 0

    # stopping the tracer
    tr.stop_saving_sessions()
    _simple_bql_query(bdb)
    assert get_num_entries(bdb.execute) == num

    # restarting the tracer
    tr.start_saving_sessions()
    _simple_bql_query(bdb)
    assert get_num_entries(bdb.execute) > num

def test_sessions_json_dump():
    (bdb, tr) = make_bdb_with_sessions()
    _simple_bql_query(bdb)
    tr.stop_saving_sessions()
    json_str = tr.dump_current_session_as_json()
    session = json.loads(json_str)
    assert isinstance(session, dict)
    assert 'version' in session
    assert session['version'] == __version__
    assert 'fields' in session
    assert isinstance(session['fields'], list)
    nfields = len(session['fields'])
    assert 'entries' in session
    assert isinstance(session['entries'], list)
    entries = session['entries']
    assert len(entries) == get_num_entries(bdb.execute)
    for entry in entries:
        assert len(entry) == nfields

def _nonexistent_table_helper(executor, tr):
    query = 'SELECT * FROM nonexistent_table'
    try:
        executor(query)
        assert False
    except apsw.SQLError:
        #tr._start_new_session()
        assert tr._check_error_entries(tr.session_id) > 0

def test_sessions_error_entry_bql():
    (bdb, tr) = make_bdb_with_sessions()
    _nonexistent_table_helper(bdb.execute, tr)

def test_sessions_error_entry_sql():
    (bdb, tr) = make_bdb_with_sessions()
    _nonexistent_table_helper(bdb.sql_execute, tr)

def test_sessions_no_errors():
    with test_core.analyzed_bayesdb_generator(test_core.t1(),
            10, None, max_seconds=1) as (bdb, generator_id):
        tr = sescap.SessionOrchestrator(bdb)
        # simple query
        cursor = bdb.execute('''
            SELECT age, weight FROM t1
                WHERE label = 'frotz'
                ORDER BY weight
        ''')
        cursor.fetchall()
        # add a metamodel and do a query
        cursor = bdb.execute('''
            ESTIMATE PREDICTIVE PROBABILITY OF age FROM t1_cc
        ''')
        cursor.fetchall()
        # there should be no error entries in the previous session
        #tr._start_new_session()
        assert tr._check_error_entries(tr.session_id) == 0

class Boom(Exception): pass
class ErroneousMetamodel(troll.TrollMetamodel):
    def __init__(self):
        self.call_ct = 0
    def name(self): return 'erroneous'
    def logpdf_joint(self, *_args, **_kwargs):
        if self.call_ct > 10: # Wait to avoid raising during sqlite's prefetch
            raise Boom()
        self.call_ct += 1
        return 0

def test_sessions_error_metamodel():
    with test_core.t1() as (bdb, _generator_id):
        bayeslite.bayesdb_register_metamodel(bdb, ErroneousMetamodel())
        bdb.execute('''
            CREATE GENERATOR t1_err FOR t1
                USING erroneous(age NUMERICAL)
        ''')
        tr = sescap.SessionOrchestrator(bdb)
        cursor = bdb.execute('''
            ESTIMATE PREDICTIVE PROBABILITY OF age FROM t1_err
        ''')
        with pytest.raises(Boom):
            cursor.fetchall()
        #tr._start_new_session()
        assert tr._check_error_entries(tr.session_id) > 0

def test_sessions_send_data():
    class response(object):
        @property
        def text(self):
            return 'Session uploaded. Thank you!'
    was_called = defaultdict(int)
    def post(url, data):
        assert url == 'https://projects.csail.mit.edu/probcomp/bayesdb/save_sessions.cgi'
        assert isinstance(data, dict)
        assert 2 == len(data)
        assert 'session_json' in data
        assert isinstance(data['session_json'], str)
        was_called['post'] += 1
        return response()
    (bdb, tr) = make_bdb_with_sessions(post=post)
    _simple_bql_query(bdb)
    tr.send_session_data()
    assert 1 == was_called['post']

def test_sessions_send_data__ci_network():
    (bdb, tr) = make_bdb_with_sessions()
    _simple_bql_query(bdb)
    tr.send_session_data()

def test_error():
    (bdb, tr) = make_bdb_with_sessions()
    with pytest.raises(apsw.SQLError):
        bdb.execute('select x from nonexistent_table')
    assert 1 == bdb.sql_execute('''
            SELECT COUNT(*) FROM bayesdb_session_entries
                WHERE type = 'bql' AND error LIKE '%no such table%'
        ''').fetchvalue()
    assert 1 == bdb.sql_execute('''
            SELECT COUNT(*) FROM bayesdb_session_entries
                WHERE type = 'sql' AND error LIKE '%no such table%'
        ''').fetchvalue()
    assert 0 == bdb.sql_execute('''
            SELECT COUNT(*) FROM bayesdb_session_entries
                WHERE type = 'bql' AND error IS NULL
        ''').fetchvalue()
    assert 0 == bdb.sql_execute('''
            SELECT COUNT(*) FROM bayesdb_session_entries
                WHERE type = 'bql' AND end_time IS NULL
        ''').fetchvalue()



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2015, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Reading data from pandas dataframes."""

import bayeslite.core as core

from bayeslite.sqlite3_util import sqlite3_quote_name

def bayesdb_read_pandas_df(bdb, table, df, create=False, ifnotexists=False,
        index=None):
    """Read data from a pandas dataframe into a table.

    :param bayeslite.BayesDB bdb: BayesDB instance
    :param str table: name of table
    :param pandas.DataFrame df: pandas dataframe
    :param bool create: if true and `table` does not exist, create it
    :param bool ifnotexists: if true, and `create` is true` and `table`
        exists, read data into it anyway
    :param str index: name of column for index

    If `index` is `None`, then the dataframe's index dtype must be
    convertible to int64, and it is mapped to the table's rowids.  If
    the dataframe's index dtype is not convertible to int64, you must
    specify `index` to give a primary key for the table.
    """
    if not create:
        if ifnotexists:
            raise ValueError('Not creating table whether or not exists!')
    column_names = [str(column) for column in df.columns]
    if index is None:
        create_column_names = column_names
        insert_column_names = ['_rowid_'] + column_names
        try:
            key_index = df.index.astype('int64')
        except ValueError:
            raise ValueError('Must specify index name for non-integral index!')
    else:
        if index in df.columns:
            raise ValueError('Index name collides with column name: %r'
                % (index,))
        create_column_names = [index] + column_names
        insert_column_names = create_column_names
        key_index = df.index
    with bdb.savepoint():
        if core.bayesdb_has_table(bdb, table):
            if create and not ifnotexists:
                raise ValueError('Table already exists: %s' % (repr(table),))
            core.bayesdb_table_guarantee_columns(bdb, table)
            unknown = set(name for name in create_column_names
                if not core.bayesdb_table_has_column(bdb, table, name))
            if len(unknown) != 0:
                raise ValueError('Unknown columns: %s' % (list(unknown),))
        elif create:
            qccns = map(sqlite3_quote_name, create_column_names)
            def column_schema(column_name, qcn):
                if column_name == index:
                    return '%s NUMERIC PRIMARY KEY' % (qcn,)
                else:
                    return '%s NUMERIC' % (qcn,)
            schema = ','.join(column_schema(ccn, qccn)
                for ccn, qccn in zip(create_column_names, qccns))
            qt = sqlite3_quote_name(table)
            bdb.sql_execute('CREATE TABLE %s(%s)' % (qt, schema))
            core.bayesdb_table_guarantee_columns(bdb, table)
        else:
            raise ValueError('No such table: %s' % (repr(table),))
        qt = sqlite3_quote_name(table)
        qicns = map(sqlite3_quote_name, insert_column_names)
        sql = 'INSERT INTO %s (%s) VALUES (%s)' % \
            (qt, ','.join(qicns), ','.join('?' for _qicn in qicns))
        for key, i in zip(key_index, df.index):
            bdb.sql_execute(sql, (key,) + tuple(df.ix[i]))



# ChaCha core: family of PRFs from 128-bit inputs to 512-bit outputs
# with 256-bit key, defined in
#
#       Daniel J. Bernstein, `ChaCha, a variant of Salsa20', Workshop
#       Record of SASC 2008: The State of the Art of Stream Ciphers.
#       Document ID: 4027b5256e17b9796842e6d0f68b0b5e
#       http://cr.yp.to/chacha/chacha-20080128.pdf

# Copyright (c) 2015 Taylor R. Campbell
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

###############################################################################
### WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING ###
###############################################################################
###                                                                         ###
###   This code is NOT constant-time -- using it makes you vulnerable to    ###
###   timing side channels.  Do not use this for cryptographic purposes.    ###
###                                                                         ###
###############################################################################
### WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING ###
###############################################################################

# Even on CPUs that do provide constant-time 32-bit bitwise operations
# and addition, CPython has no way to reliably take advantage of them.
# For example, on 32-bit CPUs, CPython uses a different representation
# of integers < 2^31 and integers >= 2^31, with significantly different
# timings for arithmetic.
#
# This is unlike C, in which carefully written programs are unlikely to
# have timing side channels even though nothing prohibits compilers
# from compiling them to run in variable-time (and to be safe you
# should check the assembly and confirm that the microarchitecture
# executing the instructions will execute them in constant time and so
# on).  In contrast, Python practically guarantees that all arithmetic
# *will* have timing side channels.

import struct

def quarterround(x, a, b, c, d):
    # a += b; d ^= a; d <<<= 16
    x[a] = (x[a] + x[b]) & 0xffffffff
    x[d] ^= x[a]
    x[d] = ((x[d] & 0xffff) << 16) | (x[d] >> 16)
    # c += d; b ^= c; b <<<= 12
    x[c] = (x[c] + x[d]) & 0xffffffff
    x[b] ^= x[c]
    x[b] = ((x[b] & 0xfffff) << 12) | (x[b] >> 20)
    # a += b; d ^= a; d <<<= 8
    x[a] = (x[a] + x[b]) & 0xffffffff
    x[d] ^= x[a]
    x[d] = ((x[d] & 0xffffff) << 8) | (x[d] >> 24)
    # c += d; b ^= c; b <<<= 7
    x[c] = (x[c] + x[d]) & 0xffffffff
    x[b] ^= x[c]
    x[b] = ((x[b] & 0x1ffffff) << 7) | (x[b] >> 25)

def quarterround_selftest():
    # From https://tools.ietf.org/html/draft-nir-cfrg-chacha20-poly1305-00
    x = [0x11111111, 0x01020304, 0x9b8d6f43, 0x01234567]
    quarterround(x, 0, 1, 2, 3)
    if x != [0xea2a92f4, 0xcb1cf8ce, 0x4581472e, 0x5881c4bb]:
        raise Exception('chacha quarterround 0 1 2 3 self-test failed')
    x = [
        0x879531e0, 0xc5ecf37d, 0x516461b1, 0xc9a62f8a,
        0x44c20ef3, 0x3390af7f, 0xd9fc690b, 0x2a5f714c,
        0x53372767, 0xb00a5631, 0x974c541a, 0x359e9963,
        0x5c971061, 0x3d631689, 0x2098d9d6, 0x91dbd320,
    ]
    quarterround(x, 2, 7, 8, 13)
    if x != [
            0x879531e0, 0xc5ecf37d, 0xbdb886dc, 0xc9a62f8a,
            0x44c20ef3, 0x3390af7f, 0xd9fc690b, 0xcfacafd2,
            0xe46bea80, 0xb00a5631, 0x974c541a, 0x359e9963,
            0x5c971061, 0xccc07c79, 0x2098d9d6, 0x91dbd320,
    ]:
        raise Exception('chacha quarterround 2 7 8 13 self-test failed')
quarterround_selftest()

def core(n, o, i, k, c):
    assert n%2 == 0
    x = [
        c[0], c[1], c[2], c[3], k[0], k[1], k[2], k[3],
        k[4], k[5], k[6], k[7], i[0], i[1], i[2], i[3],
    ]
    y = x[0:16]                 # allow o = k
    for r in range(n/2):
        quarterround(y, 0, 4, 8,12)
        quarterround(y, 1, 5, 9,13)
        quarterround(y, 2, 6,10,14)
        quarterround(y, 3, 7,11,15)
        quarterround(y, 0, 5,10,15)
        quarterround(y, 1, 6,11,12)
        quarterround(y, 2, 7, 8,13)
        quarterround(y, 3, 4, 9,14)
    o[0] = (x[0] + y[0]) & 0xffffffff
    o[1] = (x[1] + y[1]) & 0xffffffff
    o[2] = (x[2] + y[2]) & 0xffffffff
    o[3] = (x[3] + y[3]) & 0xffffffff
    o[4] = (x[4] + y[4]) & 0xffffffff
    o[5] = (x[5] + y[5]) & 0xffffffff
    o[6] = (x[6] + y[6]) & 0xffffffff
    o[7] = (x[7] + y[7]) & 0xffffffff
    o[8] = (x[8] + y[8]) & 0xffffffff
    o[9] = (x[9] + y[9]) & 0xffffffff
    o[10] = (x[10] + y[10]) & 0xffffffff
    o[11] = (x[11] + y[11]) & 0xffffffff
    o[12] = (x[12] + y[12]) & 0xffffffff
    o[13] = (x[13] + y[13]) & 0xffffffff
    o[14] = (x[14] + y[14]) & 0xffffffff
    o[15] = (x[15] + y[15]) & 0xffffffff

const32 = struct.unpack('<IIII', 'expand 32-byte k')
if const32 != (0x61707865, 0x3320646e, 0x79622d32, 0x6b206574):
    raise Exception('chacha constant32 self-test failed')

def core_selftest():
    # From http://tools.ietf.org/html/draft-strombergson-chacha-test-vectors-00
    o = [0] * 16
    i = [0] * 4
    k = [0] * 8
    core(8, o, i, k, const32)
    if o != list(struct.unpack('<IIIIIIIIIIIIIIII', bytes(bytearray([
            0x3e, 0x00, 0xef, 0x2f, 0x89, 0x5f, 0x40, 0xd6,
            0x7f, 0x5b, 0xb8, 0xe8, 0x1f, 0x09, 0xa5, 0xa1,
            0x2c, 0x84, 0x0e, 0xc3, 0xce, 0x9a, 0x7f, 0x3b,
            0x18, 0x1b, 0xe1, 0x88, 0xef, 0x71, 0x1a, 0x1e,
            0x98, 0x4c, 0xe1, 0x72, 0xb9, 0x21, 0x6f, 0x41,
            0x9f, 0x44, 0x53, 0x67, 0x45, 0x6d, 0x56, 0x19,
            0x31, 0x4a, 0x42, 0xa3, 0xda, 0x86, 0xb0, 0x01,
            0x38, 0x7b, 0xfd, 0xb8, 0x0e, 0x0c, 0xfe, 0x42,
    ])))):
        raise Exception('chacha8 core self-test failed')
    core(20, o, i, k, const32)
    if o != list(struct.unpack('<IIIIIIIIIIIIIIII', bytes(bytearray([
            0x76, 0xb8, 0xe0, 0xad, 0xa0, 0xf1, 0x3d, 0x90,
            0x40, 0x5d, 0x6a, 0xe5, 0x53, 0x86, 0xbd, 0x28,
            0xbd, 0xd2, 0x19, 0xb8, 0xa0, 0x8d, 0xed, 0x1a,
            0xa8, 0x36, 0xef, 0xcc, 0x8b, 0x77, 0x0d, 0xc7,
            0xda, 0x41, 0x59, 0x7c, 0x51, 0x57, 0x48, 0x8d,
            0x77, 0x24, 0xe0, 0x3f, 0xb8, 0xd8, 0x4a, 0x37,
            0x6a, 0x43, 0xb8, 0xf4, 0x15, 0x18, 0xa1, 0x1c,
            0xc3, 0x87, 0xb6, 0x69, 0xb2, 0xee, 0x65, 0x86,
    ])))):
        raise Exception('chacha20 core self-test failed')
    # From https://tools.ietf.org/html/draft-nir-cfrg-chacha20-poly1305-00
    i = struct.unpack('<IIII', bytes(bytearray([
        0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x09,
        0x00, 0x00, 0x00, 0x4a, 0x00, 0x00, 0x00, 0x00,
    ])))
    k = struct.unpack('<IIIIIIII', bytes(bytearray([
        0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
        0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
        0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
        0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
    ])))
    core(20, o, i, k, const32)
    if o != [
       0xe4e7f110, 0x15593bd1, 0x1fdd0f50, 0xc47120a3,
       0xc7f4d1c7, 0x0368c033, 0x9aaa2204, 0x4e6cd4c3,
       0x466482d2, 0x09aa9f07, 0x05d7c214, 0xa2028bd9,
       0xd19c12b5, 0xb94e16de, 0xe883d0cb, 0x4e3c50a2,
    ]:
        raise Exception('chacha20 core self-test failed')
core_selftest()



# Pseudorandom number generator using ChaCha8.  This does NOT provide
# key erasure / forward secrecy / backtracking resistance.

# Copyright (c) 2015 Taylor R. Campbell
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

import chacha
import os
import struct
import threading

class WeakPRNG(object):
    const32 = chacha.const32
    zero = [0,0,0,0, 0,0,0,0]

    def __init__(self, seed):
        self.key = struct.unpack('<IIIIIIII', seed)
        self.ctr = [0,0,0,0]
        self.buf = [0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0]

    def weakrandom32(self):
        # We abuse buf[15] as a decrementing counter for the number
        # 32-bit words remaining to be yielded, since we refill the
        # buffer only when we need at least one word.
        buf = self.buf
        if buf[15] == 0:
            ctr = self.ctr
            chacha.core(8, buf, ctr, self.key, self.const32)
            t = 1
            for i in range(4):  # 4 is overkill
                t = ctr[i] + t
                ctr[i] = t & 0xffffffff
                t >>= 32
            r = buf[15]
            buf[15] = 15
            return r
        buf[15] -= 1
        return buf[buf[15]]

    def weakrandom64(self):
        rh = self.weakrandom32()
        rl = self.weakrandom32()
        return (rh << 32) | rl

    def weakrandom_uniform(self, n):
        assert 0 < n
        nbits = n.bit_length()
        nwords = (nbits + 31)/32
        l = ((1 << nbits) - n) % n
        while True:
            r = 0
            for i in range(nwords):
                r <<= 32
                r |= self.weakrandom32()
            if r < l:
                continue
            return (r % n)

    def weakrandom_bytearray(self, buf, start, end):
        assert end <= len(buf)
        assert start <= end
        nbytes = end - start
        if nbytes < 128:
            nwords = nbytes/4
            for i in range(nwords):
                buf[start + 4*i : start + 4*(i + 1)] = \
                    struct.pack('<I', self.weakrandom32())
            nextra = nbytes - 4*nwords
            if 0 < nextra:
                buf[start + 4*nwords : start + 4*nwords + nextra] = \
                    struct.pack('<I', self.weakrandom32())[0 : nextra]
        else:
            subkey = [self.weakrandom32() for i in range(8)]
            const32 = self.const32
            ctr = [0,0,0,0]
            out = [0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0]
            nblocks = nbytes/64
            for i in range(nblocks):
                chacha.core(8, out, ctr, subkey, const32)
                # Overflows after 2^64 blocks.  Not worth carrying.
                t = ctr[0] + 1; ctr[0] = t & 0xffffffff; ctr[1] += t
                buf[start + 64*i : start + 64*(i + 1)] = \
                    struct.pack('<IIIIIIIIIIIIIIII', *out)
            nextra = nbytes - 64*nblocks
            if 0 < nextra:
                chacha.core(8, out, ctr, subkey, const32)
                buf[start + 64*nblocks : start + 64*nblocks + nextra] = \
                    struct.pack('<IIIIIIIIIIIIIIII', *out)[0 : nextra]
            subkey[0:8] = out[0:8] = out[8:16] = self.zero

    def weakrandom_bytes(self, n):
        buf = bytearray(n)
        self.weakrandom_bytearray(buf, 0, n)
        return bytes(buf)

def weakprng(seed):
    return WeakPRNG(seed)

if weakprng(bytes(bytearray([0] * 32))).weakrandom64() != 0x42fe0c0eb8fd7b38:
    raise Exception('weakprng self-test failed')

nondeterministic_weakprng_local = threading.local()
nondeterministic_weakprng_local.prng = None

def nondeterministic_weakprng():
    if nondeterministic_weakprng_local.prng == None:
        nondeterministic_weakprng_local.prng = weakprng(os.urandom(32))
    return nondeterministic_weakprng_local.prng

def nondeterministic_weakrandom32():
    return nondeterministic_weakprng().weakrandom32()
def nondeterministic_weakrandom64():
    return nondeterministic_weakprng().weakrandom64()
def nondeterministic_weakrandom_uniform(n):
    return nondeterministic_weakprng().weakrandom_uniform(n)
def nondeterministic_weakrandom_bytearray(buf, start, end):
    return nondeterministic_weakprng().weakrandom_bytearray(buf, start, end)
def nondeterministic_weakrandom_bytes(n):
    return nondeterministic_weakprng().weakrandom_bytes(n)

nondeterministic_weakrandom32()
nondeterministic_weakrandom64()
nondeterministic_weakrandom_uniform(123)
nondeterministic_weakrandom_bytearray(bytearray([0] * 16), 3, 8)
nondeterministic_weakrandom_bytearray(bytearray([0] * 80), 3, 67)
nondeterministic_weakrandom_bytearray(bytearray([0] * 160), 3, 132)
nondeterministic_weakrandom_bytearray(bytearray([0] * 1024), 3, 1023)
nondeterministic_weakrandom_bytearray(bytearray([0] * 1024), 3, 963)
nondeterministic_weakrandom_bytes(42)



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2015, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Metamodel interface.

To be used to model data in a :class:`bayeslite.BayesDB` handle, a metamodel
must first be registered with :func:`bayesdb_register_metamodel`.

The Crosscat metamodel is registered by default, but we can suppress
that for illustration::

   import bayeslite
   import crosscat.LocalEngine
   from bayeslite.metamodels.crosscat import CrosscatMetamodel

   bdb = bayeslite.bayesdb_open(pathname='foo.bdb', builtin_metamodels=False)
   cc = crosscat.LocalEngine.LocalEngine(seed=0)
   bayeslite.bayesdb_register_metamodel(bdb, CrosscatMetamodel(cc))

Then you can model a table with Crosscat and query the probable
implications of the data in the table::

   bdb.execute('create generator t_cc for t using crosscat(guess(*))')
   bdb.execute('initialize 10 models for t_cc')
   bdb.execute('analyze t_cc for 10 iterations wait')
   for x in bdb.execute('estimate pairwise dependence probablity from t_cc'):
       print x
"""

import bayeslite.core as core

builtin_metamodels = []
builtin_metamodel_names = set()

def bayesdb_builtin_metamodel(metamodel):
    name = metamodel.name()
    assert name not in builtin_metamodel_names
    builtin_metamodels.append(metamodel)
    builtin_metamodel_names.add(name)

def bayesdb_register_builtin_metamodels(bdb):
    """Register all builtin metamodels in `bdb`."""
    for metamodel in builtin_metamodels:
        bayesdb_register_metamodel(bdb, metamodel)

def bayesdb_register_metamodel(bdb, metamodel):
    """Register `metamodel` in `bdb`, creating any necessary tables.

    `metamodel` must not already be registered in any BayesDB, nor any
    metamodel by the same name.
    """
    name = metamodel.name()
    if name in bdb.metamodels:
        raise ValueError('Metamodel already registered: %s' % (name,))
    with bdb.savepoint():
        metamodel.register(bdb)
        bdb.metamodels[name] = metamodel

def bayesdb_deregister_metamodel(bdb, metamodel):
    """Deregister `metamodel`, which must have been registered in `bdb`."""
    name = metamodel.name()
    assert name in bdb.metamodels
    assert bdb.metamodels[name] == metamodel
    del bdb.metamodels[name]

class IBayesDBMetamodel(object):
    """BayesDB metamodel interface.

    Subclasses of :class:`IBayesDBMetamodel` implement the
    functionality needed by probabilistic BQL queries to sample from
    and inquire about the posterior distribution of a generative model
    conditioned on data in a table.  Instances of subclasses of
    `IBayesDBMetamodel` contain any in-memory state associated with
    the metamodel in the database.
    """

    def name(self):
        """Return the name of the metamodel as a str."""
        raise NotImplementedError

    def register(self, bdb):
        """Install any state needed for the metamodel in `bdb`.

        Called by :func:`bayeslite.bayesdb_register_metamodel`.

        Normally this will create SQL tables if necessary.
        """
        raise NotImplementedError

    def create_generator(self, bdb, table, schema, instantiate):
        """Create a generator for a table with the given schema.

        Called when executing ``CREATE GENERATOR``.

        Must parse `schema` to determine the column names and
        statistical types of the generator, and then call
        `instantiate` with a list of ``(column_name, stattype)``
        pairs.  `instantiate` will return a generator id and a list of
        ``(colno, column_name, stattype)`` triples.

        The generator id and column numbers may be used to create
        metamodel-specific records in the database for the generator
        with foreign keys referring to the ``bayesdb_generator`` and
        ``bayesdb_generator_column`` tables.

        `schema` is a list of schema items corresponding to the
        comma-separated ‘columns’ from a BQL ``CREATE GENERATOR``
        command.  Each schema item is a list of strings or lists of
        schema items, corresponding to whitespace-separated tokens and
        parenthesized lists.  Note that within parenthesized lists,
        commas are not excluded.
        """
        raise NotImplementedError

    def drop_generator(self, bdb, generator_id):
        """Drop any metamodel-specific records for a generator.

        Called when executing ``DROP GENERATOR``.
        """
        raise NotImplementedError

    def rename_column(self, bdb, generator_id, oldname, newname):
        """Note that a table column has been renamed.

        Not currently used.  To be used in the future when executing::

            ALTER TABLE <table> RENAME COLUMN <oldname> TO <newname>
        """
        raise NotImplementedError

    def initialize_models(self, bdb, generator_id, modelnos, model_config):
        """Initialize the specified model numbers for a generator."""
        raise NotImplementedError

    def drop_models(self, bdb, generator_id, modelnos=None):
        """Drop the specified model numbers of a generator.

        If none are specified, drop all models.
        """
        raise NotImplementedError

    def analyze_models(self, bdb, generator_id, modelnos=None, iterations=1,
            max_seconds=None, ckpt_iterations=None, ckpt_seconds=None):
        """Analyze the specified model numbers of a generator.

        If none are specified, analyze all of them.

        :param int iterations: maximum number of iterations of analysis for
            each model
        :param int max_seconds: requested maximum number of seconds to analyze
        :param int ckpt_iterations: number of iterations before committing
            results of analysis to the database
        :param int ckpt_seconds: number of seconds before committing results of
            anlaysis to the database
        """
        raise NotImplementedError

    def column_dependence_probability(self, bdb, generator_id, modelno, colno0,
            colno1):
        """Compute ``DEPENDENCE PROBABILITY OF <col0> WITH <col1>``."""
        raise NotImplementedError

    def column_mutual_information(self, bdb, generator_id, modelno, colno0,
            colno1, numsamples=100):
        """Compute ``MUTUAL INFORMATION OF <col0> WITH <col1>``."""
        raise NotImplementedError

    def row_similarity(self, bdb, generator_id, modelno, rowid, target_rowid,
            colnos):
        """Compute ``SIMILARITY TO <target_row>`` for given `rowid`."""
        raise NotImplementedError

    def predict(self, bdb, generator_id, modelno, colno, rowid, threshold,
            numsamples=None):
        """Predict a value for a column, if confidence is high enough."""
        value, confidence = self.predict_confidence(bdb, generator_id, modelno,
            colno, rowid, numsamples=numsamples)
        if confidence < threshold:
            return None
        return value

    def predict_confidence(self, bdb, generator_id, modelno, colno, rowid,
            numsamples=None):
        """Predict a value for a column and return confidence."""
        raise NotImplementedError

    def simulate_joint(self, bdb, generator_id, targets, constraints, modelno,
            num_predictions=1):
        """Simulate `targets` from a generator, subject to `constraints`.

        Returns a list of lists of values for the specified targets.

        `modelno` may be `None`, meaning "all models"

        `targets` is a list of ``(rowid, colno)`` pairs.

        `constraints` is a list of ``(rowid, colno, value)`` triples.

        `num_predictions` is the number of results to return.

        The results are samples from the distribution on targets,
        independent conditioned on (the latent state of the metamodel
        and) the constraints.
        """
        raise NotImplementedError

    def logpdf_joint(self, bdb, generator_id, targets, constraints,
            modelno=None):
        """Evalute the joint probability of `targets` subject to `constraints`.

        Returns the probability density of the targets (in log domain).

        `targets` is a list of ``(rowid, colno, value)`` triples.

        `constraints` is a list of ``(rowid, colno, value)`` triples.

        `modelno` is a model number or `None`, meaning all models.
        """
        raise NotImplementedError

    def insertmany(self, bdb, generator_id, rows):
        """Insert `rows` into a generator, updating analyses accordingly.

        `rows` is a list of tuples with one value for each column
        modelled by the generator.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2015, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# If some modules are not found, we use others, so no need to warn:
# pylint: disable=import-error
try:
    from setuptools import setup
    from setuptools.command.build_py import build_py
    from setuptools.command.sdist import sdist
    from setuptools.command.test import test
except ImportError:
    from distutils.core import setup
    from distutils.cmd import Command
    from distutils.command.build_py import build_py
    from distutils.command.sdist import sdist

    class test(Command):
        def __init__(self, *args, **kwargs):
            Command.__init__(self, *args, **kwargs)
        def initialize_options(self): pass
        def finalize_options(self): pass
        def run(self): self.run_tests()
        def run_tests(self): Command.run_tests(self)
        def set_undefined_options(self, opt, val):
            Command.set_undefined_options(self, opt, val)

def get_version():
    with open('VERSION', 'rb') as f:
        version = f.read().strip()

    # Append the Git commit id if this is a development version.
    if version.endswith('+'):
        tag = 'v' + version[:-1]
        try:
            import subprocess
            desc = subprocess.check_output([
                'git', 'describe', '--dirty', '--match', tag,
            ])
        except Exception:
            version += 'unknown'
        else:
            assert desc.startswith(tag)
            import re
            match = re.match(r'v([^-]*)-([0-9]+)-(.*)$', desc)
            if match is None:       # paranoia
                version += 'unknown'
            else:
                ver, rev, local = match.groups()
                version = '%s.post%s+%s' % (ver, rev, local.replace('-', '.'))
                assert '-' not in version

    return version

version = get_version()

def write_version_py(path):
    try:
        with open(path, 'rb') as f:
            version_old = f.read()
    except IOError:
        version_old = None
    version_new = '__version__ = %r\n' % (version,)
    if version_old != version_new:
        print 'writing %s' % (path,)
        with open(path, 'wb') as f:
            f.write(version_new)

def sha256_file(pathname):
    import hashlib
    sha256 = hashlib.sha256()
    with open(pathname, 'rb') as source_file:
        for block in iter(lambda: source_file.read(65536), ''):
            sha256.update(block)
    return sha256

def uptodate(path_in, path_out, path_sha256):
    import errno
    try:
        with open(path_sha256, 'rb') as file_sha256:
            # Strip newlines and compare.
            if file_sha256.next()[:-1] != sha256_file(path_in).hexdigest():
                return False
            if file_sha256.next()[:-1] != sha256_file(path_out).hexdigest():
                return False
    except (IOError, OSError) as e:
        if e.errno != errno.ENOENT:
            raise
        return False
    return True

def commit(path_in, path_out, path_sha256):
    import os
    with open(path_sha256 + '.tmp', 'wb') as file_sha256:
        file_sha256.write('%s\n' % (sha256_file(path_in).hexdigest(),))
        file_sha256.write('%s\n' % (sha256_file(path_out).hexdigest(),))
    os.rename(path_sha256 + '.tmp', path_sha256)

def generate_parser(lemonade, path_y):
    import distutils.spawn
    import os.path
    root = os.path.dirname(os.path.abspath(__file__))
    lemonade = os.path.join(root, *lemonade.split('/'))
    base, ext = os.path.splitext(path_y)
    assert ext == '.y'
    path_py = base + '.py'
    path_sha256 = base + '.sha256'
    if uptodate(path_y, path_py, path_sha256):
        return
    print 'generating %s -> %s' % (path_y, path_py)
    distutils.spawn.spawn([
        '/usr/bin/env', 'PYTHONPATH=' + lemonade,
        lemonade + '/bin/lemonade',
        '-s',                   # Write statistics to stdout.
        path_y,
    ])
    commit(path_y, path_py, path_sha256)

class local_build_py(build_py):
    def run(self):
        write_version_py(version_py)
        for grammar in grammars:
            generate_parser(lemonade, grammar)
        build_py.run(self)

# XXX For inexplicable reasons, during sdist.run, setuptools quietly
# modifies self.distribution.metadata.version to replace plus signs by
# hyphens -- even where they are explicitly allowed by PEP 440.
# distutils does not do this -- only setuptools.
class local_sdist(sdist):
    # This is not really a subcommand -- it's actually a predicate to
    # determine whether to run a subcommand.  So modifying anything in
    # it is a little evil.  But it'll do.
    def fixidioticegginfomess(self):
        self.distribution.metadata.version = version
        return False
    sub_commands = [('sdist_fixidioticegginfomess', fixidioticegginfomess)]

# XXX Several horrible kludges here to make `python setup.py test' work:
#
# - Standard setputools test command searches for unittest, not
#   pytest.
#
# - pytest suggested copypasta assumes . works in sys.path; we
#   deliberately make . not work in sys.path and require ./build/lib
#   instead, in order to force a clean build.
#
# - Must set PYTHONPATH too for shell tests, which fork and exec a
#   subprocess which inherits PYTHONPATH but not sys.path.
#
# - build command's build_lib variable is relative to source
#   directory, so we must assume os.getcwd() gives that.
class local_test(test):
    def __init__(self, *args, **kwargs):
        test.__init__(self, *args, **kwargs)
        self.test_suite = ' '.join(test_directories)
        self.build_lib = None
    def finalize_options(self):
        test.finalize_options(self)
        # self.build_lib = ...
        self.set_undefined_options('build', ('build_lib', 'build_lib'))
    def run_tests(self):
        import pytest
        import sys
        sys.path = [os.path.join(os.getcwd(), self.build_lib)] + sys.path
        os.environ['BAYESDB_WIZARD_MODE'] = '1'
        os.environ['BAYESDB_DISABLE_VERSION_CHECK'] = '1'
        os.environ['PYTHONPATH'] = ':'.join(sys.path)
        sys.exit(pytest.main(test_directories))

# XXX These should be attributes of `setup', but helpful distutils
# doesn't pass them through when it doesn't know about them a priori.
version_py = 'src/version.py'
lemonade = 'external/lemonade/dist'
grammars = [
    'src/grammar.y',
]
test_directories = [
    'shell/tests',
    'tests',
]

setup(
    name='bayeslite',
    version=version,
    description='BQL database built on SQLite3',
    url='http://probcomp.csail.mit.edu/bayesdb',
    author='MIT Probabilistic Computing Project',
    author_email='bayesdb@mit.edu',
    license='Apache License, Version 2.0',
    install_requires=[
        'bayeslite-apsw>=3.8.0',
        'crosscat>=0.1.48',
        'numpy',
        'requests',
    ],
    tests_require=[
        'pandas',
        'pexpect',
        'pytest',
    ],
    packages=[
        'bayeslite',
        'bayeslite.metamodels',
        'bayeslite.plex',
        'bayeslite.shell',
        'bayeslite.weakprng',
    ],
    package_dir={
        'bayeslite': 'src',
        'bayeslite.plex': 'external/plex/dist/Plex',
        'bayeslite.shell': 'shell/src',
        'bayeslite.weakprng': 'external/weakprng/dist',
    },
    # Not in this release, perhaps later.
    #scripts=['shell/scripts/bayeslite'],
    cmdclass={
        'build_py': local_build_py,
        'sdist': local_sdist,
        'test': local_test,
    },
)



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2015, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pytest
import tempfile

from bayeslite import bayesdb_open
from bayeslite import bayesdb_upgrade_schema
from bayeslite.exception import BayesDBException
from bayeslite.schema import bayesdb_schema_required

import test_core

def test_schema_upgrade():
    with bayesdb_open(version=6) as bdb:
        bayesdb_schema_required(bdb, 6, 'test pre-upgrade 6')
        with pytest.raises(BayesDBException):
            bayesdb_schema_required(bdb, 7, 'test pre-upgrade 7')
        test_core.t1_schema(bdb)
        test_core.t1_data(bdb)
        bayesdb_upgrade_schema(bdb)
        bayesdb_schema_required(bdb, 6, 'test post-upgrade 6')
        bayesdb_schema_required(bdb, 7, 'test post-upgrade 7')
        with pytest.raises(BayesDBException):
            # Nobody'll ever bump the schema version this many times,
            # right?
            bayesdb_schema_required(bdb, 1000000, 'test a gazillion')

def test_schema_incompatible():
    with tempfile.NamedTemporaryFile(prefix='bayeslite') as f:
        with bayesdb_open(pathname=f.name, version=6) as bdb:
            bayesdb_schema_required(bdb, 6, 'test incompatible 0/6')
            with pytest.raises(BayesDBException):
                bayesdb_schema_required(bdb, 7, 'test incompatible 0/7')
        with bayesdb_open(pathname=f.name) as bdb:
            bayesdb_schema_required(bdb, 6, 'test incompatible 1/6')
            bayesdb_schema_required(bdb, 7, 'test incompatible 1/7')
            bayesdb_upgrade_schema(bdb)
        with bayesdb_open(pathname=f.name) as bdb:
            bayesdb_schema_required(bdb, 6, 'test incompatible 2/6')
            bayesdb_schema_required(bdb, 7, 'test incompatible 2/7')

def test_schema_compatible():
    with tempfile.NamedTemporaryFile(prefix='bayeslite') as f:
        with bayesdb_open(pathname=f.name, version=6) as bdb:
            bayesdb_schema_required(bdb, 6, 'test compatible 0/6')
            with pytest.raises(BayesDBException):
                bayesdb_schema_required(bdb, 7, 'test compatible 0/7')
        with bayesdb_open(pathname=f.name, compatible=True) as bdb:
            bayesdb_schema_required(bdb, 6, 'test nonautomatic 1/6')
            with pytest.raises(BayesDBException):
                bayesdb_schema_required(bdb, 7, 'test compatible 1/7')
            bayesdb_upgrade_schema(bdb)
        with bayesdb_open(pathname=f.name, compatible=True) as bdb:
            bayesdb_schema_required(bdb, 6, 'test compatible 2/6')
            bayesdb_schema_required(bdb, 7, 'test compatible 2/7')
